/*
   RS_ESP32_Monitor.ino  V1.0
   Compiled for Dfrobot Firebeetle 2 ESP32-E using ESP32 board support V3.0.3

   by Matthew Ford,  2021/12/06
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/
// buzzer and mp3 msg version

#include <WiFiClient.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "httpWindowStateTask.h"
#include "WindowOpenClosedBad.h"
#include "SafeString.h"
#include "SafeStringReader.h"
#include "millisDelay.h"
#include <WiFi.h>
#include "ESP32_WS2812Flasher.h"
#include "ESPAutoWiFiConfig.h"
#include "ESPBufferedClient.h"
#include "pfodLinkedPointerList.h"
#include "locks.h"

// uncomment this line to make the buzzer buzz for 3sec on startup to test it.
//#define BUZZER_TESTING

// uncomment this line to play msg on startup for testing
// #define MSG_TESTING

// uncomment this line to output debug msgs
//#define DEBUG

// connection details for Rain Sensor and WindowOpen Detector
const char* Audio_RainSensor_host = "10.1.1.201"; // Audio Rain Detector IP (telnet)
const char* WindowOpenDetector_host_path = "http://10.1.1.202/status"; // Window Open Bridge IP (http)

// The webpage is built in msg SaftString, increase its size to fit the webpage for your installation
// uncomment DEBUG above to see any msg overflow errors.
cSF(msg, 1024); // webpage length nominally 850 for the 5 windows in this project


AsyncWebServer server(80);
const uint16_t port = 23;

typedef enum RAIN_STATE_ENUM { DRY = 0, RAIN = 1, UNKNOWN = 2} rainStateEnum;

static Stream *debugPtr = NULL;
//#define LOG_RAIN_DETECTOR
void startWebServer();

int Mp3SpeakerPin = 15; // GPIO pin number == A4 on Dfrobot Firebeetle 2 ESP32-E
int BuzzerTonePin = 14; // GPIO pin number == D6 on Dfrobot Firebeetle 2 ESP32-E
const int RGB_Pin = 5; // WS2812 RGB led GPIO5 on Firebeetle ESP32-E has RGB led on GPIO5 so use 5 for led
ESP32_WS2812Flasher LedControl = ESP32_WS2812Flasher(RGB_Pin); // default no invert, GREEN
bool highForLedOn = true;

volatile rainStateEnum rainMonitorState = UNKNOWN;
volatile rainStateEnum rainDetectorState = UNKNOWN;
volatile bool startRaining = false;
volatile bool stopRaining = false;

int failedConnectCounter = 0;
const int MAX_FAILED_CONNECTS = 10; // reboot if get this many telnet connection fails

bool httpGetFailed = false; // set true in 10 or more window state gets fail

pfodLinkedPointerList<WindowOpenClosedBad> listOfWindows;

millisDelay dryDelay;
unsigned long DRY_DELAY_ms = 1000; // only 1sec delay ,  for 10mins use 10ul * 60 * 1000; // do not switch to DRY state until no rain for this many ms
// this prevents multiple triggers when rain intermittent.

millisDelay rebootDelay;
unsigned long REBOOT_DELAY_ms = 24ul * 60 * 60 * 1000; // 24hrs in mS
millisDelay failedConnectRebootDelay;
unsigned long FAILED_CONNECT_REBOOT_DELAY_ms = 5ul * 60 * 1000; // reboot in 5mins this leave webpage etc running most of the time

millisDelay buzzerDelay;
unsigned long BUZZER_DELAY_ms = 200; // toggle tone to intermittent
millisDelay buzzerOnTimer;
unsigned long BUZZER_ON_TIMER_ms = 3000; // on for 3sec  only play buzzer for 3sec each time it starts raining and a window is open

const unsigned long MSG_LENGTH_ms = 18000; // length of msg
millisDelay msgDelay;
// use the MSG_DELAY to adjust time between repeats of msg when raining and window open
unsigned long MSG_DELAY_ms = MSG_LENGTH_ms + 30000; // delay between repeats of msg  30sec
millisDelay msgTriggerDelay;
unsigned long MSG_TRIGGER_DELAY_ms = 100; // pulse P1 low for this long to trigger/retrigger msg


volatile bool msgRunning = false;
// call this to trigger msg, can call this repeatedly
void triggerMsg() {
  if (msgRunning) {
    return;
  }
  msgRunning = true;
  msgTriggerDelay.start(MSG_TRIGGER_DELAY_ms);
  digitalWrite(Mp3SpeakerPin, LOW);
}

// call this each every loop() to control msg
void updateMsg() {
  if (!msgRunning) {
    return;
  }
  if (msgTriggerDelay.justFinished()) {
    digitalWrite(Mp3SpeakerPin, HIGH);
    msgDelay.start(MSG_DELAY_ms);
    return;
  }
  if (msgDelay.justFinished()) {
    msgRunning = false; // can retrigger now
  }
}

void initMp3Speaker() {
  pinMode(Mp3SpeakerPin, OUTPUT);
  digitalWrite(Mp3SpeakerPin, HIGH);
}

volatile bool buzzerOn = false;
bool isBuzzerOn() {
  return buzzerOn;
}

volatile bool buzzerDisabled = true; // start with disable on restart, need not raining or reenable
void disableBuzzer() {
  buzzerDisabled = true;
  setBuzzer(false);
}

bool isBuzzerDisabled() {
  return buzzerDisabled;
}

void initBuzzer() {
  pinMode(BuzzerTonePin, OUTPUT);
  digitalWrite(BuzzerTonePin, LOW);
  buzzerOn = false;
}

void setBuzzer(bool _on) {
  static bool tone = true;
  if (buzzerOn != _on) {
    // state change
    buzzerOn = _on;
    if (buzzerOn) { // turn on
      tone = true;
      buzzerDelay.start(BUZZER_DELAY_ms);
      digitalWrite(BuzzerTonePin, HIGH);
    } else { // turn off
      buzzerDelay.stop();
      digitalWrite(BuzzerTonePin, LOW);
    }
    return;
  }
  if (!buzzerOn) {
    return;
  }
  // else on
  if (buzzerDelay.justFinished()) {
    buzzerDelay.repeat();
    tone = !tone;
    if (tone) {
      digitalWrite(BuzzerTonePin, HIGH);
    } else {
      digitalWrite(BuzzerTonePin, LOW);
    }
  }
}

void buzzerUpdate() {
  setBuzzer(buzzerOn);
}

WiFiClient rainDetectorClient; // for connection
millisDelay rainDetectorReconnectDelay;
const unsigned long RECONNECT_DELAY_ms = 5000; // 5sec
millisDelay rainDetectorkeepAliveTimer;
const unsigned long KEEP_ALIVE_TIMER_ms = 5000; // 5sec

// returns true if connected else false
bool connectClientToHost(WiFiClient &client, const char* host, const uint16_t port) {
  if (!client.connect(host, port)) {
    if (debugPtr) {
      debugPtr->print("Connection to "); debugPtr->print(host); debugPtr->print(":"); debugPtr->print(port); debugPtr->println(" failed.");
    }
    return false;
  }
  return true;
}

void markRainDetectorUnkown() {
  rainDetectorState = UNKNOWN;
}

volatile unsigned long uptimeDays = 0;
volatile unsigned long uptimeHrs = 0;
volatile unsigned long lastChangeDays = 0;
volatile unsigned long lastChangeHrs = 0;
volatile unsigned long uptimeMins = 0;
volatile unsigned long lastChangeMins = 0;

void secToHrsMins(unsigned long sec, volatile unsigned long &days, volatile unsigned long &hrs, volatile unsigned long &mins) {
  unsigned long m = sec / 60;
  hrs = m / 60;
  mins = m % 60;
  days = hrs / 24;
  hrs = hrs % 24;
}

createSafeStringReader(sfReaderRainDetector, 120, "\n\r");
void handleRainDetectorClient() {
  if (!rainDetectorClient.connected()) {
    if (rainDetectorkeepAliveTimer.isRunning()) {
      rainDetectorkeepAliveTimer.stop();
      sfReaderRainDetector.end();
      rainDetectorClient.stop();
    }
    if (!rainDetectorReconnectDelay.isRunning()) {
      rainDetectorReconnectDelay.start(RECONNECT_DELAY_ms);
    }
  }
  if (rainDetectorReconnectDelay.justFinished()) {
    if (connectClientToHost(rainDetectorClient, Audio_RainSensor_host, port)) {
      rainDetectorClient.println();
      rainDetectorClient.setTimeout(5);
      rainDetectorkeepAliveTimer.start(KEEP_ALIVE_TIMER_ms);
      sfReaderRainDetector.connect(rainDetectorClient);
    } else {
      failedConnectCounter++;
      rainDetectorClient.stop();
      markRainDetectorUnkown();
    }
  }
  if (rainDetectorClient.connected()) {
    if (sfReaderRainDetector.read()) {
      sfReaderRainDetector.trim();
      rainStateEnum lastRainState = rainDetectorState;
#ifdef LOG_RAIN_DETECTOR
      Serial.println(sfReaderRainDetector);
#endif
      if (sfReaderRainDetector.indexOf("Dry") >= 0) {
        rainDetectorState = DRY;
      } else if (sfReaderRainDetector.indexOf("Rain") >= 0) {
        rainDetectorState = RAIN;
      } else {
        if (debugPtr) {
          debugPtr->print("Unknown data: "); debugPtr->println(sfReaderRainDetector);
        }
      }
      if (lastRainState != rainDetectorState) { // state changed
        if (rainDetectorState == DRY) {
          if (lastRainState == RAIN) {
            stopRaining = true;
          }
          if (debugPtr) {
            debugPtr->print(" Dry  ");
            debugPtr->println();
          }
        } else if (rainDetectorState == RAIN) {
          if (lastRainState == DRY) {
            startRaining = true;
          }
          if (debugPtr) {
            debugPtr->print(" Rain ");
            debugPtr->println();
          }
        } else { // bad
        }
      }
      // parse last change and current counter
      cSF(sfField, 30); // temp SafeString to received fields, max field len is <30;
      char delims[] = ","; // fields delimited by , or *
      bool returnEmptyFields = true; // return empty field for ,,
      int idx = 0;
      idx = sfReaderRainDetector.stoken(sfField, idx, delims, returnEmptyFields);
      // first field is state change count
      bool countsOK = true;
      unsigned long lastChangeCount = 0;
      if (!sfField.toUnsignedLong(lastChangeCount)) {
        if (debugPtr) {
          debugPtr->print("invalid last change count: "); debugPtr->println(sfField);
        }
        countsOK = false;
      }
      idx = sfReaderRainDetector.stoken(sfField, idx, delims, returnEmptyFields); // dry/rain
      idx = sfReaderRainDetector.stoken(sfField, idx, delims, returnEmptyFields); // current count
      unsigned long currentCount = 0;
      if (!sfField.toUnsignedLong(currentCount)) {
        if (debugPtr) {
          debugPtr->print("invalid current count: "); debugPtr->println(sfField);
        }
        countsOK = false;
      }
      if (countsOK) {
        unsigned long uptimeSec = currentCount / 20;
        secToHrsMins(uptimeSec, uptimeDays, uptimeHrs, uptimeMins);
        unsigned long lastChangeSec = (currentCount - lastChangeCount) / 20;
        secToHrsMins(lastChangeSec, lastChangeDays, lastChangeHrs, lastChangeMins);
      }
      if (debugPtr) {
        debugPtr->print(" Uptime "); debugPtr->print(uptimeHrs); debugPtr->print(":");
        if (uptimeMins < 10) {
          debugPtr->print('0');
        }
        debugPtr->print(uptimeMins); debugPtr->print("");
        debugPtr->print(" Has been "); debugPtr->print(rainDetectorState == RAIN ? "Raining" : "Dry"); debugPtr->print(" for ");
        debugPtr->print(lastChangeHrs); debugPtr->print(":");
        if (lastChangeMins < 10) {
          debugPtr->print('0');
        }
        debugPtr->print(lastChangeMins); debugPtr->print("  (hh:mm)");
        debugPtr->println();
      }
    }
    if (rainDetectorkeepAliveTimer.justFinished()) {
      rainDetectorkeepAliveTimer.restart();
      rainDetectorClient.println();
    }
  }
}

bool windowOpen() {
  START_LOCK(windowListLock);
  WindowOpenClosedBad *devicePtr = listOfWindows.getFirst();
  while (devicePtr) {
    if (devicePtr->getState() == OPEN) { // ignore BAD
      return true;
    } // else
    devicePtr = listOfWindows.getNext();
  }
  END_LOCK(windowListLock);
  return false;
}

void setBADonConnectionFail() {
  START_LOCK(windowListLock);
  WindowOpenClosedBad *devicePtr = listOfWindows.getFirst();
  while (devicePtr) {
    devicePtr->setState(BAD);
    devicePtr = listOfWindows.getNext();
  }
  END_LOCK(windowListLock);
}

// returns NULL if not found
WindowOpenClosedBad* getWindowOpenClosedBad(SafeString & name) {
  WindowOpenClosedBad *devicePtr = NULL;
  START_LOCK(windowListLock);
  devicePtr = listOfWindows.getFirst();
  while (devicePtr) {
    if (name == (devicePtr->getDeviceName())) {
      break;
    } // else
    devicePtr = listOfWindows.getNext();
  }
  END_LOCK(windowListLock);
  return devicePtr; // null not found
}

// returns NULL if not found
void printWindowOpenClosedBad(SafeString & name) {
  if (!debugPtr) {
    return;
  }
  debugPtr->println(" -- compared ---  ");
  debugPtr->print("'"); debugPtr->print(name); debugPtr->println("'");
  START_LOCK(windowListLock);
  WindowOpenClosedBad *devicePtr = listOfWindows.getFirst();
  debugPtr->println(" -- to ---  ");
  while (devicePtr) {
    debugPtr->print("'"); debugPtr->print(devicePtr->getDeviceName()); debugPtr->print("'  ");
    if (name == (devicePtr->getDeviceName())) {
      debugPtr->println(" matched ");
    } else {
      debugPtr->println(" NO match ");
    }
    devicePtr = listOfWindows.getNext();
  }
  END_LOCK(windowListLock);
  debugPtr->println(" ---------  ");
}

// only add devices if initFlag is true, else skip those not already there
void processWindowOpenLine(SafeString & sfName) {
  if (sfName.isEmpty()) {
    return;
  }
  if (debugPtr) {
    debugPtr->print("Device name: ");
    debugPtr->println(sfName);
  }
  // only scan for == upto first ,
  // look at names with trailing ,O or ,C
  if ((!sfName.endsWith(",O")) && (!sfName.endsWith(",C")) && (!sfName.endsWith(",B"))) {
    return; // skip this one
  } // else
  windowStateEnum windowState = BAD;
  if (sfName.endsWith(",O")) {
    windowState = OPEN;
  } else if (sfName.endsWith(",C")) {
    windowState = CLOSED;
  }

  int idx = sfName.indexOf(',');
  if (idx >= 0) {
    sfName.substring(sfName, 0, idx);
  }
  sfName.trim();
  START_LOCK(windowListLock);
  WindowOpenClosedBad *devicePtr = getWindowOpenClosedBad(sfName);
  if ((!devicePtr)) { // not found
    // else if initializing add
    // not  found add it upto first ,
    if (debugPtr) {
      debugPtr->print("Adding Device name: "); debugPtr->println(sfName);
      printWindowOpenClosedBad(sfName);
    }
    devicePtr = new WindowOpenClosedBad(sfName.c_str()); // note MUST use new since pfodLinkedPointerList uses delete when remove() called
    listOfWindows.add(devicePtr);
  }
  // update lastseen
  windowStateEnum lastState = devicePtr->getState();
  devicePtr->setState(windowState); // save window state
  if (lastState != windowState) {
    devicePtr->setLastChanged(millis());
  }
  END_LOCK(windowListLock);
}


void setup() {
  initBuzzer();
  Serial.begin(115200);

#ifdef DEBUG
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  SafeString::setOutput(Serial); // for debug error msgs
  debugPtr = &Serial;
  debugPtr->println();
#endif
  initMp3Speaker();

  LedControl.setColor(0, 0, 0);
  LedControl.setOnOff(PIN_OFF);

  //  /* Initialise wifi module */
  // setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
  if (ESPAutoWiFiConfigSetup(-RGB_Pin, highForLedOn, 0)) { // negative pin number indicates RGB led, Firebeetle ESP32-E has RGB led on GPIO5 so use -5 for led
    return; // in config mode so skip rest of setup
  }

  if (initializeLocks() != 0) { // reboot here will flash green led slowly
    while (1); // reboot
  }
  startWebServer();
  setHttpWindowStateTaskDebugOut(debugPtr);
  createHttpWindowStateTask(WindowOpenDetector_host_path);

  rebootDelay.start(REBOOT_DELAY_ms); // start reboot timer
  enableLoopWDT(); // default appears to be 5sec

#ifdef MSG_TESTING
  triggerMsg();
#endif

#ifdef BUZZER_TESTING
  // buzzer testing
  buzzerOnTimer.start(BUZZER_ON_TIMER_ms);
  setBuzzer(true);  // sound until first rain sensor reading
#endif
  if (debugPtr) {
    debugPtr->print(" Dry  ");
    debugPtr->println();
  }
}


void loop() {
  wifiConnected_v = false;
  if (ESPAutoWiFiConfigLoop()) {  // handle auto configure/connect
    return;
  }
  wifiConnected_v = true; // ok connected / reconnected

  // else normal loop() code
  if (rebootDelay.justFinished()) {
    while (1) {} // force watch dog timer reboot
  }

  bool windowIsOpen = windowOpen(); // check once per loop()

  if (windowIsOpen && (rainDetectorState == RAIN)) {
    triggerMsg(); // msg will restart after MSG_DELAY_ms
  }
  updateMsg(); // handle msg timers

  if (windowIsOpen) {
    LedControl.setColor(166, 35, 0); // Orange
    LedControl.setOnOff(PIN_ON);
  } else {
    LedControl.setOnOff(PIN_OFF);
  }

  if (!isHttpWindowStateConnectionGood()) {
    httpGetFailed = true;
  } else {
    httpGetFailed = false;
  }

  buzzerUpdate(); // update buzzer
  if (startRaining) {
    startRaining = false;
    dryDelay.stop();
    if (rainMonitorState != RAIN) {  // buzz on rain start if was not raining already
      rainMonitorState = RAIN;
      if (windowIsOpen) {
        if (!buzzerOnTimer.isRunning()) { // not already buzzing
          buzzerOnTimer.start(BUZZER_ON_TIMER_ms);
          setBuzzer(true);
        }
      }
    }
  }
  if (buzzerOnTimer.justFinished()) {
    setBuzzer(false);  // turn off buzzer
  }
  if (stopRaining) {
    stopRaining = false;
    dryDelay.start(DRY_DELAY_ms);
  }
  if (dryDelay.justFinished()) {
    rainMonitorState = DRY;     // switch to DRY if no rain for DRY_DELAY_ms
  }

  handleRainDetectorClient();

  // check for failed connections
  if (rainDetectorClient.connected()) {
    failedConnectCounter = 0; //reset
    failedConnectRebootDelay.stop();
  }
  if (failedConnectCounter > MAX_FAILED_CONNECTS) {
    failedConnectRebootDelay.start(FAILED_CONNECT_REBOOT_DELAY_ms);
  }
  if (failedConnectRebootDelay.justFinished()) {
    while (1) {} // force watch dog timer reboot
  }
}

//// ========== trivial async webserver example =============
void notFound(AsyncWebServerRequest *request) {
  request->send(404, "text/plain", "Not found");
}


void handleRoot(AsyncWebServerRequest *request) {
  if (debugPtr) {
    debugPtr->println(">>> WebServer handleRoot");
  }
  unsigned long secSinceReboot = millis() / 1000;
  unsigned long minsSinceReboot;
  unsigned long hrsSinceReboot;
  unsigned long daysSinceReboot;
  secToHrsMins(secSinceReboot, daysSinceReboot, hrsSinceReboot, minsSinceReboot);

  msg.clear();
  msg =
    "<html>\
  <head>\
    <meta http-equiv='refresh' content='5'/>\
    <title>Rain and Window Open Monitor</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>";
  msg += "<h1>";
  msg += "Rain and Window Open Monitor";
  msg += "<font size=\"-1\">";
  msg += "<br>";
  msg += "Monitor restarted ";
  if (hrsSinceReboot > 0) {
    msg += hrsSinceReboot;
    msg += " hrs ";
  }
  msg += minsSinceReboot;
  msg += " mins ago";
  msg += "</font>";
  msg += "</h1> ";
  msg += "<b><font size=\"+2\">";

  if (rainDetectorState == UNKNOWN) {
    msg += "<b><font color=\"red\">";
    msg += "No connection to Rain Detector";
    msg += "</font>";
  } else {
    if (rainDetectorState == DRY) {
      msg += "Dry for ";
    } else {
      msg += "Raining for ";
    }
    if (lastChangeDays > 0) {
      msg += lastChangeDays;
      msg += " days ";
    }
    if (lastChangeHrs > 0) {
      msg += lastChangeHrs;
      msg += " hrs ";
    }
    msg += lastChangeMins;
    msg += " mins";
  }
  msg += "</font>";

  msg += "<br>";
  msg += "<br>";
  msg += "The BLE Window Open devices found were: -<br>";
  // loop through and add reading found
  START_LOCK(windowListLock);
  WindowOpenClosedBad *devicePtr = listOfWindows.getFirst();
  //  msg += "<h1>";
  if (httpGetFailed) {
    msg += "<b><font size=\"+2\" color=\"red\">";
    msg += "Lost connection to Window Open Montor.<br>";
    msg += "</font></b>";
  } else {
    if (devicePtr == NULL) {
      msg += "No window monitors found so far.<br>";
    } else {
      while (devicePtr != NULL) {  // normal
        const char *deviceName = devicePtr->getDeviceName();
        windowStateEnum windowState = devicePtr->getState();
        unsigned long lastWindowChangedSec = (millis() - devicePtr->getLastChanged()) / 1000;
        unsigned long lastWindowChangedMins = 0;
        unsigned long lastWindowChangedHrs = 0;
        unsigned long lastWindowChangedDays = 0;
        secToHrsMins(lastWindowChangedSec, lastWindowChangedDays, lastWindowChangedHrs, lastWindowChangedMins);

        if (windowState == BAD) { // missed two advert interval 10+90 x 2
          msg += "<font size=\"+1\" color=\"black\">";
          msg += "<b>";
          msg += " BAD    -- ";
          msg += deviceName;
          msg += " - not advertising. Battery Low or Out Of Range";
          msg += "</b>";
        } else if (windowState == OPEN) {
          msg += "<font size=\"+2\" color=\"red\">";
          msg += "<b>";
          msg += " OPEN   -- ";
          msg += deviceName;
          msg += "</b>";
        } else {
          msg += "<font color=\"green\">";
          msg += " CLOSED -- ";
          msg += deviceName;
        }
        msg += " for ";
        if (lastWindowChangedDays > 0) {
          msg += lastWindowChangedDays;
          msg += " days ";
        }
        if (lastWindowChangedHrs > 0) {
          msg += lastWindowChangedHrs;
          msg += " hrs ";
        }
        msg += lastWindowChangedMins;
        msg += " mins";

        msg += "</font>";
        msg += "<br>";
        devicePtr = listOfWindows.getNext();
      }
    }
  }
  END_LOCK(windowListLock);
  msg += "</body></html>";
  request->send(200, "text/html", msg.c_str());
}

void startWebServer() {
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    handleRoot(request);
  });

  server.onNotFound(notFound);
  server.begin();
  if (debugPtr) {
    debugPtr->println("webserver started");
  }
}
